-------------------------------- MODULE MPGC --------------------------------

(*  A specification that factorizes the common structure 
    of fast and safe modes *)

EXTENDS RDR, Library

CONSTANTS Initial, Acceptor

\*  RespQuorum is the set of quorums used to determine a response
\*  AbortQuorum is the set of quorums used to determine an abort value
\*  AbortValues([Q -> S]) is the set of safe abort values given 
\*  the dStates of a quorum Q of acceptors.
CONSTANTS RespQuorum, AbortQuorum, AbortValues(_)


VARIABLES status, pending, initVals, dState, accStatus, interface, 
  pastPending
\* abortVals is a history variable
VARIABLE abortVals

INSTANCE SpecLinInterface

vars == <<status, pending, initVals, dState, accStatus, interface, 
  abortVals, pastPending>>
Labels == {"idle", "ready", "pending", "panic", "aborted"}
AcceptorLabels == {"idle","ready","stopped"}

TypeInvariant ==
  /\ \forall p \in P :
    /\ status[p] \in Labels
    /\ pending[p] \in Req
  /\ \forall r \in Acceptor :
    /\ dState[r] \in S
    /\ accStatus[r] \in AcceptorLabels
  /\ initVals = {}
  /\ pastPending \subseteq Req

Init ==
  /\ status = 
    IF Initial
    THEN [p\in P |-> "ready"]
    ELSE  [p\in P |-> "idle"]
  /\ pending = [p \in P |-> Some(Req)]
  /\ initVals = {}
  /\ dState = [r \in Acceptor |-> Bot]
  /\ accStatus = 
    IF Initial
    THEN [r \in Acceptor |-> "ready"]
    ELSE [r \in Acceptor |-> "idle"]
  /\ interface = InterfaceInit
  /\ abortVals = {}
  /\ pastPending = {}

Ini(p) == \E c \in C, v \in S :
  /\ status[p] = "idle"
  /\ pending' = [pending EXCEPT ![p] = <<p,c>>]
  /\ initVals' =  initVals \cup {v}
  /\ status' = [status EXCEPT ![p] = "pending"]
  /\ Initialize(p, c, v)
  /\ pastPending' = pastPending \cup {<<p,c>>}
  /\ UNCHANGED <<dState, accStatus, abortVals>>

Inv(p) == \E c \in C :
  /\ status[p] = "ready"
  /\ pending' = [pending EXCEPT ![p] = <<p,c>>]
  /\ status' = [status EXCEPT ![p] = "pending"]
  /\ Invoke(p, c)
  /\ pastPending' = pastPending \cup {<<p,c>>}
  /\ UNCHANGED <<dState, accStatus, initVals, abortVals>>

SrvStates(Q) == 
  {s \in S : \E srv \in Q : s = dState[srv]}

Res(p) == 
  /\ status[p] = "pending"
  /\ status' = [status EXCEPT ![p] = "ready"]
  /\ \E Q \in RespQuorum :
    /\ \A srv \in Q : accStatus[srv] # "idle"
    /\ LET glb == GLB(SrvStates(Q))
       IN /\  Contains(glb, pending[p])
          /\  Response(p, Output(glb, pending[p]))
  /\ UNCHANGED <<pending, initVals, dState, accStatus, 
    abortVals, pastPending>>

(*  Computing an abort value:
    all thirds contain at least one RDR of the Quorum that was used for
    the last commit. Therefore every GLB is either a prefix of the last
    committed RDR or an extension of it with pending requests. *)
Abo(p) ==
  /\ status[p] = "panic"
  /\ \E Q \in AbortQuorum :
    /\ \A srv \in Q : accStatus[srv] = "stopped"
    /\ \E s \in AbortValues([a \in Q |-> dState[a]]) : 
      /\ Abort(p, pending[p][2], s)
      /\ abortVals' = abortVals \cup {s} 
  /\ status' = [status EXCEPT ![p] = "aborted"]
  /\ UNCHANGED <<pending, initVals, dState, accStatus, pastPending>>

\* We abstract over time: a process can panic at any moment.
Panic(p) ==
  /\ status[p] = "pending"
  /\ status' = [status EXCEPT ![p] = "panic"]
  /\ UNCHANGED <<pending, initVals, dState, accStatus, interface, 
    abortVals, pastPending>>

Pending == 
  {p \in P : status[p] \in {"pending","panic","aborted"}}

\* A Acceptor executes a pending request.
Exec(r) ==
  /\ accStatus[r] = "ready"
  /\ \E req \in pastPending : 
    dState' = [dState EXCEPT ![r] = @ \bullet req]
  /\ UNCHANGED <<status, pending, initVals, accStatus, interface, 
    abortVals, pastPending>>

\* A Acceptor sets its local state to one of the init values of the processes.
WakeUp(r) ==
  /\ accStatus[r] = "idle"
  /\ \E iv \in initVals : 
    /\  dState' = [dState EXCEPT ![r] = iv]
    /\  accStatus' = [accStatus EXCEPT ![r] = "ready"]
  /\ UNCHANGED <<status, pending, initVals, interface, abortVals, pastPending>>

Stop(r) == \E p \in P :
    /\  status[p] \in {"panic","aborted"}
    /\  accStatus[r] = "ready"
    /\  accStatus' = [accStatus EXCEPT ![r] = "stopped"]
    /\  UNCHANGED <<status, pending, initVals, interface, dState, 
      abortVals, pastPending>>

Next ==
    \/  \E p \in P : Ini(p) \/ Inv(p) \/ Res(p) \/ Abo(p) \/ Panic(p)
    \/  \E r \in Acceptor : Exec(r) \/ WakeUp(r) \/ Stop(r)

Spec == Init /\ [][Next]_vars

=============================================================================
\* Modification History
\* Last modified Sun Dec 08 19:43:47 CET 2013 by losa
\* Created Fri Nov 29 14:12:15 CET 2013 by losa
