------------------------------ MODULE Quorum ------------------------------

EXTENDS Generic, Library, TLCDefs
INSTANCE RDR

CONSTANTS AbortQuorum, RespQuorum, Initial, Acceptor

INSTANCE FastMPGCDefs

VARIABLES status, pending, initVals, execAcks, panicAcks, dState, 
  network, accStatus, interface
\*  abortVals and pastPending are history variables
VARIABLE abortVals, pastPending

INSTANCE SpecLinInterface

vars == <<status, pending, initVals, execAcks, panicAcks, dState, 
  accStatus, interface, network, abortVals, pastPending>>

Labels == {"idle", "ready", "pending", "panic", "aborted"}
AcceptorLabels == {"idle","ready","stopped"}

Agent == P \cup Acceptor

Msg == {<<"req", r>> : r \in Req} \cup {<<"execAck",s>> : s \in S}
  \cup {<<"panic">>} \cup {<<"panicAck",s>> : s\in S}
  \cup {<<"init", s>> : s \in S}
INSTANCE Network

TypeInvariant ==
  /\ \forall p \in P :
    /\ status[p] \in Labels
    /\ pending[p] \in Req
    /\ \A a \in Acceptor : 
      /\ execAcks[p][a] \in {{s} : s \in S} \cup {{}}
      /\ panicAcks[p][a] \in {{s} : s \in S} \cup {{}}
  /\ \forall a \in Acceptor :
     /\  dState[a] \in S
     /\  accStatus[a] \in AcceptorLabels
  /\ initVals \subseteq S
  /\ abortVals \subseteq S
  /\ pastPending \subseteq Req

(*  The processes *)

InitProcs ==
  /\  status = [p \in P |-> IF Initial THEN "ready" ELSE "idle"]
  /\  pending = [p \in P |-> Some(Req)]
  /\  execAcks = [p \in P |-> [a \in Acceptor |-> {}]]
  /\  panicAcks = [p \in P |-> [a \in Acceptor |-> {}]]
    
Inv(p) ==
  /\ status[p] = "ready"
  /\ \E c \in C : 
    /\ Invoke(p, c)
    /\ Snd(p, [a \in Acceptor |-> {<<"req", <<p,c>> >>}])
    /\ pending' = [pending EXCEPT ![p] = <<p,c>>]
    /\ pastPending' = pastPending \cup {<<p,c>>}
  /\ status' = [status EXCEPT ![p] = "pending"]
  /\ UNCHANGED <<initVals, execAcks, panicAcks, dState, accStatus, 
    abortVals>>

Ini(p) ==
  /\  status[p] = "idle"
  /\ \E c \in C, s \in S : 
    /\ Initialize(p, c, s)
    /\ Snd(p, [a \in Acceptor |-> {<<"init",s>>, <<"req", <<p,c>> >>}])
    /\ pending' = [pending EXCEPT ![p] = <<p,c>>]
    /\ initVals' = initVals \cup {s}
    /\ pastPending' = pastPending \cup {<<p,c>>}
  /\ status' = [status EXCEPT ![p] = "pending"]
  /\ UNCHANGED <<execAcks, panicAcks, dState, accStatus, abortVals>>
    
RcvExecAcC(p) ==
  /\ status[p] \in {"pending","panic"}
  /\ \E s \in S, a \in Acceptor : 
    /\  Rcv(p, <<"execAck",s>>, a)
    /\  execAcks' = [execAcks EXCEPT ![p] = [@ EXCEPT ![a] = {s}]] 
  /\ UNCHANGED <<status, pending, initVals, dState, accStatus, 
    interface, abortVals, panicAcks, pastPending>>
    
RcvPanicAck(p) ==
  /\ status[p] = "panic"
  /\ \E s \in S, a \in Acceptor : 
    /\  Rcv(p, <<"panicAck",s>>, a)
    /\  panicAcks' = [panicAcks EXCEPT ![p] = [@ EXCEPT ![a] = {s}]] 
  /\ UNCHANGED <<status, pending, execAcks, initVals, dState, 
    accStatus, interface, abortVals, pastPending>>
    
Panic(p) == 
  /\ status[p] = "pending"
  /\ status' = [status EXCEPT ![p] = "panic"]
  /\ Snd(p, [a \in Acceptor |-> {<<"panic">>}])
  /\ UNCHANGED <<pending, initVals, execAcks, panicAcks, dState, 
    accStatus,interface, abortVals, pastPending>>

Res(p) ==
  /\ status[p] = "pending"
  /\ \E Q \in RespQuorum :
    /\ \A a \in Q : execAcks[p][a] # {}
    /\ LET acks == {s \in S : \E a \in Q : execAcks[p][a] = {s}}
           glb == GLB(acks)
           req == pending[p] 
       IN  /\  Contains(glb, req)
           /\  Response(p, Output(glb, req))
  /\ status' = [status EXCEPT ![p] = "ready"]
  /\ execAcks' = [execAcks EXCEPT ![p] = [a \in Acceptor |-> {}]]
  /\ UNCHANGED <<pending, initVals, panicAcks, dState, accStatus, 
    network, abortVals, pastPending>>

PanicAck(p,a) ==
  CHOOSE s \in S : panicAcks[p][a] = {s}

Abo(p) ==
  /\ status[p] = "panic"
  /\ \E Q \in AbortQuorum :
    /\ \A a \in Q : panicAcks[p][a] # {}
    /\ LET acks == [a \in Q |-> PanicAck(p,a)]
       IN \E s \in AbortValues(acks) : 
            /\ Abort(p, pending[p][2], s)
            /\ abortVals' = abortVals \cup {s}
  /\ status' = [status EXCEPT ![p] = "aborted"]
  /\ UNCHANGED <<pending, initVals, execAcks, panicAcks, dState, accStatus,
    network, pastPending>>

(*  The Acceptors  *)

InitAcceptor ==
  /\ accStatus = [a \in Acceptor |-> IF Initial THEN "ready" ELSE "idle"]
  /\ dState = [a \in Acceptor |-> Bot]

WakeUp(a) ==
  /\ accStatus[a] = "idle"
  /\ accStatus' = [accStatus EXCEPT ![a] = "ready"]
  /\ \E p \in P, s \in S : 
    /\ Rcv(a, <<"init",s>>, p)
    /\ dState' = [dState EXCEPT ![a] = s]
  /\ UNCHANGED <<status, initVals, panicAcks, pending, execAcks, 
    interface, abortVals, pastPending>>

Exec(a) ==
  /\ accStatus[a] = "ready"
  /\ \E p \in P, req \in Req :
    /\ RcvSnd(a, <<"req",req>>, p, 
            [q \in {p} |-> {<<"execAck",dState[a] \bullet req>>}])
    /\ dState' = [dState EXCEPT ![a] = @ \bullet req]
    /\ dState'[a] \in S \* For TLC
  /\ UNCHANGED <<status, initVals, pending, execAcks, interface, 
    accStatus, abortVals, panicAcks, pastPending>>

Stop(a) ==
  /\ accStatus[a] = "ready"
  /\ \E p \in P : RcvSnd(a, <<"panic">>, p, 
    [q \in {p} |-> {<<"panicAck", dState[a]>>}])
  /\ accStatus' = [accStatus EXCEPT ![a] = "stopped"]
  /\ UNCHANGED <<status, initVals, pending, execAcks, interface, dState, 
    abortVals, panicAcks, pastPending>>

(*  The full spec   *)

Init ==
  /\ InitProcs
  /\ InitAcceptor
  /\ interface = InterfaceInit
  /\ network = {}
  /\ abortVals = {}
  /\ initVals = {}
  /\ pastPending = {}

Next == 
  \/ \E p \in P : Inv(p) \/ Ini(p) \/ RcvPanicAck(p) \/ RcvExecAcC(p) 
    \/ Panic(p) \/ Abo(p) \/ Res(p)
  \/ \E a \in Acceptor : WakeUp(a) \/ Exec(a) \/ Stop(a)

Spec == Init /\ [][Next]_vars
 
Fast == INSTANCE FastMPGC

THEOREM Spec => Fast!Spec

=============================================================================
\* Modification History
\* Last modified Mon Dec 09 10:08:35 CET 2013 by nano
\* Last modified Mon Dec 09 09:38:51 CET 2013 by losa
\* Created Sun Nov 24 14:21:05 CET 2013 by losa
