-------------------------------- MODULE RDR --------------------------------

(*  Specification of Recoverable Data-Type Representations *)

EXTENDS Sequences, Naturals, FiniteSets, Library

CONSTANTS S, C, O, P, _ \bullet _, Output(_,_), Bot
\*  For the efficiency of model checking, allow substitution of star, GLB, 
\*  and Contains. The properties of the constants below are asserted in
\*  ASSUME statements.
CONSTANTS _ \star _, GLB(_), Contains(_,_), _ \preceq _

\* Requests:
Req == P \times C

\* Types of \bullet and Output:
TypeOk == 
    /\ \forall s \in S, c \in Req : s \bullet c \in S
    /\ \forall s \in S, c \in Req : Output(s,c) \in O
ASSUME TypeOk

\* Execute a sequence of requests:
RECURSIVE Star(_,_,_)
Star(s,rs,i) == 
    IF Len(rs) < i THEN s
    ELSE LET s2 == s \bullet rs[i] IN Star(s2,rs,i+1)
\* Ensures that \star and Star match.
ASSUME \A s\in S, rs \in Seq(Req) : s \star rs = Star(s,rs,1)


\* Idempotence property of data types:
Idem1 == \forall s \in S : \forall r \in Req : \forall rs \in Seq(Req) : r \in Image(rs) => s \star rs = s \star Append(rs,r)
Idem2 == \forall s \in S : \forall o \in O : \forall p,q \in P : \forall c1,c2 \in C : 
    LET r1 == <<p,c1>>
        r2 == <<q,c2>>
    IN
        Output(s,r1) = o /\ p # q 
            =>  LET s2 == (s \bullet r1) \bullet r2 
                IN  Output(s2,r1) = o
Idem == Idem1 /\ Idem2
ASSUME Idem

\* The partial order:
PrecEq(s1, s2) == 
    \/  s1 = s2
    \/  \exists rs \in Seq(Req) : s2 = s1 \star rs
ASSUME \A s1,s2 \in S : (s1 \preceq s2) = PrecEq(s1,s2)

\*Antisymmetry of RDRs
AntiSym == \forall s1,s2 \in S : s1 \preceq s2 /\ s2 \preceq s1 => s1 = s2
ASSUME AntiSym

\* Greatest lower bounds:
IsLB(s, s1, s2) == s \preceq s1 /\ s \preceq s2
IsGLB(s, s1, s2) == 
    /\ IsLB(s, s1 ,s2) 
    /\ \forall s3 \in S : s # s3 /\ IsLB(s3, s1 ,s2) => s3 \preceq s
\*Semi lattice property of RDRs:
s1 \sqcup s2 == CHOOSE s \in S : IsGLB(s, s1, s2)

GLBExists == \forall s1,s2 \in S : IsGLB(s1 \sqcup s2, s1, s2)
ASSUME GLBExists

\*GLB of a set of states:
RECURSIVE GLB1(_)
GLB1(ss) ==
    LET s ==  CHOOSE s \in ss : TRUE
    IN
        IF Cardinality(ss) = 1
        THEN s
        ELSE (s \sqcup GLB1(ss \ {s}))
ASSUME \A ss \in SUBSET S : GLB1(ss) = GLB(ss)

\* The consistency property of RDRs:
Consistency == \forall s0,s1,s2 \in S, rs1,rs2 \in Seq(Req) :
    LET rset == Image(rs1)\cup Image(rs2) 
    IN  /\  s1 = s0 \star rs1
        /\  s2 = s0 \star rs2
        =>  \exists rs \in Seq(rset) : s1 \sqcup s2 = s0 \star rs
ASSUME Consistency

\* Checking whether an RDR contains a given request:
Contains1(s,r) == \exists rs \in Seq(Req) : r \in Image(rs) /\ s = Bot \star rs
ASSUME \A s\in S, r\in Req : Contains(s,r) = Contains1(s,r)

=============================================================================
\* Modification History
\* Last modified Sun Dec 08 22:33:41 CET 2013 by losa
\* Last modified Wed Nov 13 16:23:29 CET 2013 by nano
\* Created Fri Nov 08 11:25:02 CET 2013 by losa
