------------------------------ MODULE SpecLin ------------------------------

EXTENDS Library, RDR

CONSTANT Initial \* TRUE when first instance.

VARIABLES
    status, pending, dState, initialized, abortVals, initVals, interface

INSTANCE SpecLinInterface

vars == <<status, pending, dState, interface, initVals, initialized, abortVals>>
statusStr == {"idle", "ready","aborted","pending"}

TypeInvariant == 
    /\ \A p \in P :
        /\ status[p] \in statusStr
        /\ pending[p] \in Req
    /\ dState \in S
    /\ initVals \in SUBSET S
    /\ abortVals \in SUBSET S
    
\* Initial states
Init ==
    /\  IF Initial 
        THEN    /\  status = [p \in P |-> "ready"] 
                /\  initialized = TRUE
        ELSE    /\  status = [p \in P |-> "idle"] 
                /\  initialized = FALSE
    /\ dState = Bot
    /\ pending = [p \in P |-> Some(Req)]
    /\ initVals = {}
    /\ abortVals = {}
    /\ interface = InterfaceInit
        
\* Invocation by process p:
Inv(p) == \E c \in C : 
    /\ status[p] = "ready"
    /\ status' = [status EXCEPT ![p] = "pending"]
    /\ pending' = [pending EXCEPT ![p] = <<p,c>>]
    /\ Invoke(p, c)
    /\ UNCHANGED <<dState,initialized,initVals,abortVals>>

\* Response by process p:
Resp(p) ==
    /\  status[p] = "pending"
    /\  initialized
    /\  status' = [status EXCEPT ![p] = "ready"]
    /\  Contains(dState,pending[p])
    /\  Response(p,Output(dState,pending[p]))
    /\  UNCHANGED <<dState, pending, initialized,initVals,abortVals>>

Pending == 
  {p \in P : status[p] \in {"pending", "aborted"}}
PendingReqs == 
  {pending[p] : p \in Pending}
InitSets ==
  {is \in SUBSET initVals : is # {}}

SafeInit ==
  {s1 \in S :
    /\ initVals # {}
    /\ \E is \in InitSets : 
      \E rs \in NoDupSeq1(PendingReqs) :
        s1 = GLB(is) \star rs
    /\ \A a \in abortVals : s1 \preceq a}

PossibleCommit ==
  {s1 \in S :
    /\ dState \preceq s1
    /\ \/ \E rs \in NoDupSeq1(PendingReqs) : s1 = dState \star rs
       \/ \E is \in InitSets : 
         /\ dState \preceq GLB(is)
         /\ \E rs \in NoDupSeq1(PendingReqs) : s1 = GLB(is) \star rs }

SafeCommit ==
  {s1 \in PossibleCommit :
    \A a \in abortVals : s1 \preceq a}

SafeAbort ==
  {s1 \in S :
    IF initialized 
    THEN s1 \in PossibleCommit
    ELSE \E is \in InitSets :
      \E rs \in NoDupSeq1(PendingReqs) :
        s1 = GLB(is) \star rs}

\* Abort by process p:
Abo(p) ==
  /\ status[p] = "pending" 
  /\ status' = [status EXCEPT ![p] = "aborted"]
  /\ \E s1 \in SafeAbort : 
    /\ Abort(p, pending[p][2], s1)
    /\ abortVals' = abortVals \cup {s1}
  /\ UNCHANGED <<dState, pending, initialized, initVals>>

\* Linearize some pending requests.
Lin ==
    /\ initialized
    /\ PendingReqs # {}
    /\ \E s \in SafeCommit : dState' = s
    /\ dState' \in S \* For TLC
    /\ UNCHANGED <<status, pending, interface, initialized, initVals, abortVals>>

\* Init call
Ini(p) ==
    /\ status[p] = "idle"
    /\ \E c \in C, sval \in S : 
        /\ Initialize(p, c, sval)
        /\ status' = [status EXCEPT ![p] = "pending"]
        /\ pending' = [pending EXCEPT ![p] = <<p,c>>]
        /\ initVals' = initVals \cup {sval}
    /\ UNCHANGED <<dState, initialized, abortVals>>

Recover ==
    /\ \neg initialized
    /\ \E s1 \in SafeInit : dState' = s1
    /\ dState' \in S \* For TLC
    /\ initialized' = TRUE
    /\ UNCHANGED <<pending, status, interface,initVals,abortVals>>
    
Next == \E p \in P :  Lin  \/ Inv(p) \/ Resp(p) \/ Abo(p) \/ Ini(p) \/ Recover

Spec == Init /\ [][Next]_vars

=============================================================================
\* Modification History
\* Last modified Tue Dec 03 16:39:56 CET 2013 by losa
\* Created Fri Nov 22 15:35:54 CET 2013 by losa
