--------------------------- MODULE SpecLinIsLin ---------------------------

EXTENDS Generic, Library, TLCDefs

INSTANCE RDR

VARIABLE interface

INSTANCE SpecLinInterface

Initial == TRUE \* TRUE when first instance.

VARIABLES
    status, pending, dState, initialized, abortVals, initVals
    
(*  In Linearizability one cannot linearize multiple requests at once.
    To avoid a complex prophecy variable construction we just restrict
    the Lin action.  
    We need a history variable for nxtOut and a prophecy variable to 
    introduce stuttering because in Linearizability one cannot go 
    from "ready" to "commited". 
    nxtOut is a history var and commited is a prophecy vars. *)
VARIABLES nxtOut, committed

vars == <<status, pending, dState, interface, initVals, initialized, abortVals,nxtOut,committed>>
statusStr == {"sleeping", "ready","aborted","pending"}

TypeInvariant == 
    /\ \A p \in P :
        /\ status[p] \in statusStr
        /\ pending[p] \in Req
        /\ nxtOut[p] \in O
        /\ committed[p] \in BOOLEAN
    /\ dState \in S
    /\ initVals \in SUBSET S
    /\ abortVals \in SUBSET S

\* Initial states
Init ==
    /\  IF Initial 
        THEN    /\  status = [p \in P |-> "ready"] 
                /\  initialized = TRUE
        ELSE    /\  status = [p \in P |-> "sleeping"] 
                /\  initialized = FALSE
    /\ dState = Bot
    /\ pending = [p \in P |-> Some(Req)]
    /\ initVals = {}
    /\ abortVals = {}
    /\ interface = InterfaceInit
    /\ nxtOut = [p\in P |-> Some(O)]
    /\ committed = [p\in P |-> FALSE]
        
\* Invocation by process p:
Inv(p) == \E c \in C : 
    /\ status[p] = "ready"
    /\ status' = [status EXCEPT ![p] = "pending"]
    /\ pending' = [pending EXCEPT ![p] = <<p,c>>]
    /\ Invoke(p, c)
    /\ UNCHANGED <<dState,initialized,initVals,abortVals, nxtOut, committed>>

\* Response by process p:
Resp(p) ==
    /\  status[p] = "pending"
    /\  initialized
    /\  status' = [status EXCEPT ![p] = "ready"]
    /\  Contains(dState,pending[p])
    /\  Response(p,Output(dState,pending[p]))
    /\  UNCHANGED <<dState, pending, initialized, initVals, abortVals, nxtOut>>
    /\  committed[p] = TRUE 
    /\  committed' = [committed EXCEPT ![p] = FALSE]

Pending == 
  {p \in P : status[p] \in {"pending", "aborted"}}

PendingReqs == 
  {pending[p] : p \in Pending}

InitSets ==
  {is \in SUBSET initVals : is # {}}

SafeInit ==
  {s1 \in S :
    /\ initVals # {}
    /\ \E is \in InitSets : 
      \E rs \in NoDupSeq1(PendingReqs) :
        s1 = GLB(is) \star rs
    /\ \A a \in abortVals : s1 \preceq a}
    
SafeCommit ==
  {s1 \in S :
    /\ initialized
    /\ dState \preceq s1
    /\ \E r \in PendingReqs : s1 = dState \bullet r
    /\ \A s2 \in abortVals : s1 \preceq s2}

SafeAbort ==
  {s1 \in S :
    IF initialized 
    THEN /\ dState \preceq s1
         /\ \/ \E rs \in NoDupSeq1(PendingReqs) : s1 = dState \star rs
            \/ \E is \in InitSets : 
              /\ dState \preceq GLB(is)
              /\ \E rs \in NoDupSeq1(PendingReqs) : s1 = GLB(is) \star rs
    ELSE \E is \in InitSets :
      \E rs \in NoDupSeq1(PendingReqs) :
        s1 = GLB(is) \star rs}

\* Abort by process p:
Abo(p) ==
  /\ status[p] = "pending"
  /\ status' = [status EXCEPT ![p] = "aborted"]
  /\ \E s1 \in SafeAbort : 
    /\ Abort(p, pending[p][2], s1)
    /\ abortVals' = abortVals \cup {s1}
  /\ UNCHANGED <<dState, pending, initialized, initVals, nxtOut, committed>>

\* Linearize JUST ONE pending request.
Lin ==
    /\ initialized
    /\ \E r \in PendingReqs :
      /\ dState' = dState \bullet r
      /\ committed' = [committed EXCEPT ![r[1]] = TRUE]
      /\ nxtOut' = [nxtOut EXCEPT ![r[1]] = Output(dState,pending[r[1]])]
    /\ \A a \in abortVals : dState' \preceq a
    /\ dState' \in S \* For TLC
    /\ UNCHANGED <<status, pending, interface, initialized, initVals, abortVals>>

\* Init call
Ini(p) ==
    /\ status[p] = "sleeping"
    /\ \E c \in C, sval \in S : 
        /\ Initialize(p, c, sval)
        /\ status' = [status EXCEPT ![p] = "pending"]
        /\ pending' = [pending EXCEPT ![p] = <<p,c>>]
        /\ initVals' = initVals \cup {sval}
    /\ UNCHANGED <<dState, initialized, abortVals, nxtOut, committed>>

Recover ==
    /\ \neg initialized
    /\ \E s1 \in SafeInit : dState' = s1
    /\ dState' \in S \* For TLC
    /\ initialized' = TRUE
    /\ UNCHANGED <<pending, status, interface,initVals,abortVals, nxtOut, committed>>

Commit(p) ==
    /\  status[p] = "pending"
    /\  Contains(dState,pending[p])
    /\  committed' = [committed EXCEPT ![p] = TRUE]
    /\  nxtOut' = [nxtOut EXCEPT ![p] = Output(dState,pending[p])]
    /\  UNCHANGED <<status,pending,dState,interface,initialized, initVals, abortVals>>
    
Next == 
  \E p \in P : 
    Lin  \/ Inv(p) \/ Resp(p) \/ Abo(p) \/ Ini(p) \/ Recover \/ Commit(p)

Spec == Init /\ [][Next]_vars

LinInterface == [
    inv |-> interface.inv, 
    resp |-> interface.resp]
    

L == INSTANCE Linearizability WITH
    status <- [p\in P |-> 
      IF committed[p] 
      THEN "committed" 
      ELSE 
        IF status[p] = "aborted" 
        THEN "pending" 
        ELSE status[p]],
    pending <- pending,
    dState <- dState,
    nxtOut <- nxtOut,
    interface <- LinInterface

THEOREM Spec => L!Spec
    
=============================================================================
\* Modification History
\* Last modified Fri Nov 29 20:31:55 CET 2013 by losa
\* Created Fri Nov 22 17:51:38 CET 2013 by losa
